C **********************************************************************
C
C * @author        Wolfram Ruehaak
C * @version       1.0
C * @Copyright:    GPL Copyright (c) 2003
C * @e-mail:       w.ruehaak@online.de,
C * @homepage:     http://www.geomath.onlinehome.de
C * @date:		 15. December 2003
C *
C * bln2grd: A program for conversion of a blank file into a regular 
C *          grid file.
C *          The program reads a file with a format like:
C * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
C *         Nr of data-points, 0
C *         First Datapoint x-coordinate, y-coordinate
C *         Second Datapoint x-coordinate, y-coordinate 
C *          (...),(...)
C * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
C *    Or in Numbers:
C *    4,0
C *    12,10
C *    14,18
C *    17,20
C *    17,17
C *    3,0
C *    5,4
C *    3,4
C *    3,3
C *    ... and so on ...
C * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
C *
C *    These datapoints can be polylines or polygons (then the last 
C *    point is the same as the first one).
C *
C *   These scattered datapoints are formed into a regular grid,
C *   where every cell containing a datapoint or the linking line 
C *   between two datapoints gets the value '-1', every other gets 
C *   the value '0'.
C * 
C *   If you want to transform a shape file you will need to convert it 
C *   first into a blankfile wich is easly possible for example with 
C *   program Surfer (tm) bei Golden Software.
C **********************************************************************
C *   This program is free software; you can redistribute it and/or 
C *   modify it under the terms of the GNU General Public License as 
C *   published by the Free Software Foundation; either version 2 of 
C *   the License, or (at your option) any later version.
C *
C *   This program is distributed in the hope that it will be useful,
C *   but WITHOUT ANY WARRANTY; without even the implied warranty of
C *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
C *   GNU General Public License for more details.
C *
C *   You should have received a copy of the GNU General Public License
C *   along with this program; if not, write to the Free Software
C *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  
C *   02111-1307  USA
C *
C **********************************************************************

      PROGRAM BLN2RASTER
           
      IMPLICIT NONE
      
      INCLUDE 'bln2grd.par'

      INTEGER     A, I, X, Y, FCO(NX,NY), COUNT
      INTEGER     NRI(NA)
      DOUBLE PRECISION XA(500,NA),YA(500,NA),POSX,POSY,DX,DY,M,
     & MINAX,MAXAX,MINAY,MAXAY,DELTAX,DELTAY,XS,YS,XS2,YS2,
     & DELTX2,DELTY2

      INTEGER NX0,NY0,MINX,MAXX,MINY,MAXY,MINZ,MAXZ
      
      CHARACTER FILENAME*50,FAULTDAT*100
      FAULTDAT = 'test.bln'
      FILENAME = 'test.grd'

      
      WRITE(*,*) 'BLN2RASTER WORKING'
C *********************************************************************
C     CONSTRUCT THE HYDRO BOUNDARY CONDITIONS FILE FROM THE BLANK FILE
C *********************************************************************
C
C           READ BORDERS

      WRITE(*,*) 'File with fault data = ', FAULTDAT
      OPEN(75,FILE=FAULTDAT)
      DO A = 1, NA
      READ(75,*) NRI(A)      
     
      IF(TEST) WRITE(*,*) 
     & 'A = ', A, 
     & 'NRI(',A,')=',NRI(A), 
     & 'INCLUDE LAST VALUE IN bln2grd.par'
     
            DO I = 1, NRI(A)
            READ(75,*) XA(I,A),YA(I,A)
            ENDDO
      ENDDO      
C       
C           READ 2D-GRID
C
      COUNT = 0

      POSX = XMIN
      DO X = 1, NX
      POSY = YMIN
      DO Y = 1, NY   
            DO A = 1, NA
            DO I = 1, NRI(A)-1

C there are 4 co-ordinates for the rectangle: 
C      posx, posx+fstepx, posy, posy+fstepy
C we have to solve a straight line equation for this polygon 
C     (xa(i,a),ya(i,a))-(xa(i+1,a),ya(i+1,a))
             
             MINAX = MIN(XA(I,A),XA(I+1,A))
             MAXAX = MAX(XA(I,A),XA(I+1,A))
             MINAY = MIN(YA(I,A),YA(I+1,A))
             MAXAY = MAX(YA(I,A),YA(I+1,A))

             DX = (XA(I,A) - XA(I+1,A))             
             DY = (YA(I,A) - YA(I+1,A))
                               
                  IF(DABS(DX).LE.ERR)THEN
                  M = 0.D0
                  ELSEIF(DABS(DY).LE.ERR)THEN
                  M = 0.D0
                  ELSEIF(DABS(DY).GT.ERR.AND.DABS(DX).GT.ERR)THEN
                  M = DY/DX
                  ENDIF

C     Point of intersection is in this cell when: 
C                 POSX <= (XA(I,A) + (DX * M)) <= POSX+FSTEPX
C                 POSY <= (YA(I,A) + (DY / M)) <= POSY+FSTEPY

             IF(M.NE.0.D0)THEN
C
C     add-on criteria: restriction of the area to the by  _UV_ 
C     created rectangle: 
C     _UV_ = _(XA(I,A),YA(I,A)),(XA(I+1,A),YA(I+1,A))_
C
               IF(POSX+FSTEPX.GE.MINAX.AND.POSX-FSTEPX.LE.MAXAX.AND.
     &            POSY+FSTEPY.GE.MINAY.AND.POSY-FSTEPY.LE.MAXAY)THEN

            DELTAX = POSX - MINAX
            DELTAY = POSY - MINAY
            DELTX2 = POSX - MAXAX
            DELTY2 = POSY - MAXAY

                IF( POSX        .GE.(MINAX + (DELTAX * M)).AND.
     &             (POSX+FSTEPX).LE.(MAXAX + (DELTAX * M)).OR.
     &              POSY        .GE.(MINAY + (DELTAX * M)).AND.
     &             (POSY+FSTEPY).LE.(MAXAY + (DELTAX * M)).OR.     
     &              POSY        .GE.(MINAY + (DELTAY / M)).AND.
     &             (POSY+FSTEPY).LE.(MAXAY + (DELTAY / M)).OR.
     &              POSX        .GE.(MINAX + (DELTAY / M)).AND.
     &             (POSX+FSTEPX).LE.(MAXAX + (DELTAY / M)))THEN

C      calculate point of intersection
                  XS2 = DELTAX+(.5D0*FSTEPX)
                  YS2 = DELTAY+(.5D0*FSTEPY)
                  YS = (XS2 * M)
                  XS = (YS2 / M)
      
                  IF(YS.GE.DELTAY.AND.YS.LE.DELTAY+FSTEPY.OR.
     &               XS.GE.DELTAX.AND.XS.LE.DELTAX+FSTEPX)THEN
                  FCO(X,Y) = -1
                  GOTO 100
                  ELSEIF(M.LT.0.D0)THEN                  
              IF(YS.GE.DELTY2.AND.YS.LE.DELTY2+FSTEPY.OR.
     &           XS.GE.DELTX2.AND.XS.LE.DELTX2+FSTEPX)THEN
                   FCO(X,Y) = -1
                   GOTO 100
                   ENDIF
                  ENDIF
                ELSE
                FCO(X,Y) =  0
                ENDIF
               ENDIF
              ELSE

                IF(DABS(DY).LE.ERR)THEN
C     HORIZONTAL LINE
                 IF(MINAX.LE.POSX.AND.MAXAX.GE.POSX+FSTEPX.AND.
     &              MINAY.GE.POSY.AND.MAXAY.LE.POSY+FSTEPY)THEN
                 FCO(X,Y) = -1
                 GOTO 100
                 ENDIF
                ENDIF
C ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                IF(DABS(DX).LE.ERR)THEN
C     VERTICAL LINE
                 IF(MINAY.LE.POSY.AND.MAXAY.GE.POSY+FSTEPY.AND.
     &              MINAX.GE.POSX.AND.MAXAX.LE.POSX+FSTEPX)THEN
                  FCO(X,Y) = -1
                 GOTO 100
                 ENDIF
                ENDIF
              ENDIF

C     special case xa(i,a), ya(i,a) are inside a cell
            IF(XA(I,A).GE.POSX.AND.XA(I,A).LE.POSX+FSTEPX.AND.
     &         YA(I,A).GE.POSY.AND.YA(I,A).LE.POSY+FSTEPY)THEN
            FCO(X,Y) = -1
            ELSE
            FCO(X,Y) = 0
            ENDIF
            ENDDO
            ENDDO
      
100   CONTINUE
      WRITE(99,'(I5,I5,F12.2,F12.2,I3)') 
     & X, Y, POSX, POSY, FCO(X,Y)
      IF(FCO(X,Y).LT.0) COUNT = COUNT + 1
      POSY = POSY + FSTEPY                        
      ENDDO      
      POSX = POSX + FSTEPX
      ENDDO
      
      WRITE(*,*) 'COUNT = ', COUNT
      CALL SRFAGRD_I(NX,NY,XMIN,XMAX,YMIN,YMAX,-1,0,FCO,FILENAME)
      
      RETURN
      END
